###############################################################################
###############################################################################
# PROGRAM:
# auxiliar-functions.R
# 
# USE:
# Define auxiliar functions to be used when fitting the ZMPS model
#     
# AUTHOR:
# Wesley Bertoli (wbsilva@utfpr.edu.br)
#
# DATE:
# January 22, 2021
###############################################################################
###############################################################################

# Probability mass functions

## Poisson-Sujatha distribution

dps <- function (y, a, log = F) {
  
  aux <- 3 * log(x = a) + log1p(x = y^2 + y * (a + 4) + (a^2 + 3 * a + 3)) - (y + 3) * log1p(x = a) - 
         log1p(x = a^2 + a + 1)
  
  if (log) aux else exp(x = aux)
  
}
dps <- cmpfun(f = dps)

## Zero-truncated Poisson-Sujatha distribution

dztps <- function (y, a, log = F) {
  
  aux <- 3 * log(x = a) + log1p(x = y^2 + y * (a + 4) + (a^2 + 3 * a + 3)) - y * log1p(x = a) - 
         log1p(x = a^4 + 4 * a^3 + 10 * a^2 + 7 * a + 1)
  
  if (log) aux else exp(x = aux)
  
}
dztps <- cmpfun(f = dztps)

## Zero-modified Poisson-Sujatha distribution

dzmps <- function (y, a, b, log = F) {
  
  b <- ifelse(test = b == 1, yes = 1 - 1e-12, no = ifelse(test = b == 0, yes = 1e-12, no = b))
  d <- ifelse(test = y == 0, yes = 1, no = 0)
  
  aux <- d * log1p(x = -b) + (1 - d) * (log(x = b) + dztps(y = y, a = a, log = T))
  
  if (log) aux else exp(x = aux)
  
}
dzmps <- cmpfun(f = dzmps)

###############################################################################
###############################################################################

# Mean parameterization

h <- function (a, log = F) {
  
  p1 <- 3 * sqrt(x = 3) * sqrt(x = 7 * a^4 + 28 * a^3 + 171 * a^2 + 32 * a + 5) * a
  p2 <- 8 * a^3 + 66 * a^2 + 6 * a + 1
  p  <- (p1 + p2)^(1/3)
  
  aux <- log(x = p/a - ((a - 1) * (5 * a + 1))/(a * p) - (a - 1)/a) - log(x = 3) 
  
  if (log) aux else exp(x = aux)
  
}
h <- cmpfun(f = h)

###############################################################################
###############################################################################

# Log-likelihood function of beta1

loglike.b1 <- function (a, data) {
  
  k <- which(x = data$y > 0)
  
  if (all(x = is.finite(x = a))) {
    
    e <- exp(x = h(a = exp(x = tcrossprod(x = data$X1[k,], y = t(x = a[1:ncol(x = data$X1)]))), log = T))
    s <- sum(x = dztps(y = data$y[k], a = e, log = T))
    
  }
  
  if (exists(x = 's')) s else -Inf
  
}
loglike.b1 <- cmpfun(f = loglike.b1)

# Log-likelihood function of beta2

loglike.b2 <- function (a, data) {
  
  g <- make.link(link = data$link)$linkinv
  
  if (all(x = is.finite(x = a))) {
    
    e <- tcrossprod(x = data$X2, y = t(x = a[1:ncol(x = data$X2)]))
    s <- sum(x = ifelse(test = data$y == 0, yes = 1, no = 0) * log(x = g(eta = e)^(-1) - 1)) + 
         sum(x = log(x = g(eta = e)))
    
  }
  
  if (exists(x = 's')) s else -Inf
  
}
loglike.b2 <- cmpfun(f = loglike.b2)

# Full log-likelihood function

loglike <- function (a, data) {
  
  id.b1 <- 1:ncol(x = data$X1)
  id.b2 <- (max(x = id.b1) + 1):(ncol(x = data$X1) + ncol(x = data$X2))
  
  loglike.b1(a = a[id.b1], data = data) + loglike.b2(a = a[id.b2], data = data)
  
}
loglike <- cmpfun(f = loglike)

###############################################################################
###############################################################################

# Unnormalized log-posterior distribution of beta1

logpost.b1 <- function (a, mu0, S0, data) { 
  
  if (all(x = is.finite(x = a))) { 
    
    s <- loglike.b1(a = a, data = data) + dmvnorm(x = a, mean = mu0, sigma = makePD(x = S0), log = T)
    
  }
  
  if (exists(x = 's')) s else -Inf
  
}
logpost.b1 <- cmpfun(f = logpost.b1)

# Unnormalized log-posterior distribution of beta2

logpost.b2 <- function (a, mu0, S0, data) {
  
  if (all(x = is.finite(x = a))) { 
    
    s <- loglike.b2(a = a, data = data) + dmvnorm(x = a, mean = mu0, sigma = makePD(x = S0), log = T)
    
  }
  
  if (exists(x = 's')) s else -Inf
  
}
logpost.b2 <- cmpfun(f = logpost.b2)

###############################################################################
###############################################################################

# Make positive definite matrix

makePD <- function (x, tol) {
  
  d   <- dim(x = x)[1]
  eig <- eigen(x = x)
  eiv <- eig$values
  
  if (missing(x = tol)) {
    
    tol <- d * max(x = abs(x = eiv)) * .Machine$double.eps
    
  }
  
  x + (eig$vectors %*% diag(x = pmax(x = 0, y = (2 * tol) - eiv), nrow = d) %*% t(x = eig$vectors))
  
}
makePD <- cmpfun(f = makePD)

###############################################################################
###############################################################################

# Random-walk Metropolis algorithm

RwM <- function (data, M = 5e4, burn.in = 1e4, print.it = F) {
  
  flag <- 0
  k    <- 1
  
  nu <- data$n/(data$n + 1)
  
  b1.prior <- rep(x = 0, times = ncol(x = data$X1))
  b2.prior <- rep(x = 0, times = ncol(x = data$X2))
  
  b1 <- array(dim = c(M + 1, length(x = b1.prior))) 
  b2 <- array(dim = c(M + 1, length(x = b2.prior)))
  
  b1[1,] <- b1.prior
  b2[1,] <- b2.prior
  
  prior.cov.b1 <- data$n * solve(a = crossprod(x = data$X1))
  prior.cov.b2 <- data$n * solve(a = crossprod(x = data$X2))
  
  while (k <= M) {
    
    if (print.it && k %in% floor(x = seq(from = M/10, to = M, length.out = 10))) {
      
      print(x = paste('Updating - ', k/M * 100, '%', sep = ''))
      
    }
    
    b1.star <- nu * b1[k,] + (1 - nu) * b1.prior
    b2.star <- nu * b2[k,] + (1 - nu) * b2.prior
    
    S.b1 <- solve(a = -hessian(func = logpost.b1, x = b1.star, mu0 = b1.prior, S0 = prior.cov.b1, data = data))
    S.b2 <- solve(a = -hessian(func = logpost.b2, x = b2.star, mu0 = b2.prior, S0 = prior.cov.b2, data = data))   
    
    psi1 <- as.numeric(x = mvrnorm(n = 1, mu = b1.star, Sigma = nu * makePD(x = S.b1)))
    psi2 <- as.numeric(x = mvrnorm(n = 1, mu = b2.star, Sigma = nu * makePD(x = S.b2)))
    
    a1 <- logpost.b1(a = psi1, mu0 = b1.prior, S0 = prior.cov.b1, data = data) -
          logpost.b1(a = b1[k,], mu0 = b1.prior, S0 = prior.cov.b1, data = data)
    
    a2 <- logpost.b2(a = psi2, mu0 = b2.prior, S0 = prior.cov.b2, data = data) -
          logpost.b2(a = b2[k,], mu0 = b2.prior, S0 = prior.cov.b2, data = data)
    
    b1[k + 1,] <- b1[k,]
    b2[k + 1,] <- b2[k,]
    
    u1 <- runif(n = 1)
    u2 <- runif(n = 1)
    
    if (log(x = u1) <= ifelse(test = is.finite(x = a1), yes = a1, no = -Inf) && 
        log(x = u2) <= ifelse(test = is.finite(x = a2), yes = a2, no = -Inf)) {
      
      b1[k + 1,] <- psi1
      b2[k + 1,] <- psi2
      
      flag <- flag + 1
      
    }
    
    k <- k + 1
    
  }
  
  list(chain = cbind(b1, b2)[-(1:(burn.in + 1)),], acc.rate = flag/M)
  
}
RwM <- cmpfun(f = RwM)

###############################################################################
###############################################################################

# Function to obtain estimates under the classical formulation of ZM models

or.pars <- function (chain, data) {
  
  id.b1 <- 1:ncol(x = data$X1)
  id.b2 <- (max(x = id.b1) + 1):(ncol(x = data$X1) + ncol(x = data$X2))
  
  eX <- as.numeric(x = tcrossprod(x = data$X1, y = t(x = colMeans(x = chain[, id.b1]))))
  eV <- as.numeric(x = tcrossprod(x = data$X2, y = t(x = colMeans(x = chain[, id.b2]))))
  
  mu <- exp(x = eX)
  om <- make.link(link = data$link)$linkinv(eta = eV)
  p0 <- 1 - om
  p  <- om/(1 - dps(y = 0, a = exp(x = h(a = mu, log = T))))
  
  
  c(mean(x = mu), mean(x = om), mean(x = p0), mean(x = p))
  
}
or.pars <- cmpfun(f = or.pars)

###############################################################################
###############################################################################

# Function to estimate the linear predictors

lin.pred <- function (chain, data) {
  
  id.b1 <- 1:ncol(x = data$X1)
  id.b2 <- (max(x = id.b1) + 1):(ncol(x = data$X1) + ncol(x = data$X2))
  
  eX <- eV <- array(dim = c(data$n, nrow(x = chain)))
  
  for (i in 1:nrow(x = chain)) {
    
    eX[, i] <- tcrossprod(x = data$X1, y = t(x = chain[i, id.b1]))
    eV[, i] <- tcrossprod(x = data$X2, y = t(x = chain[i, id.b2]))
    
  }
  
  list(eX = eX, eV = eV)
  
}
lin.pred <- cmpfun(f = lin.pred)

###############################################################################
###############################################################################

# Function to estimate the Conditional Predictive Ordinate (CPO)

cal.CPO <- function (M, data, lp) {
  
  CPO <- numeric(length = data$n)
  aux <- array(dim = c(data$n, M))
  
  eX <- lp$eX
  eV <- lp$eV 
  
  for (i in 1:data$n) {
    
    aux[i,] <- dzmps(y = data$y[i], a = exp(x = h(a = exp(x = eX[i,]), log = T)),
                     b = make.link(link = data$link)$linkinv(eta = eV[i,]))
    CPO[i]  <- M * (sum(x = 1/aux[i,]))^(-1)
    
  }
  
  list(CPO = CPO, aux = aux)
  
}
cal.CPO <- cmpfun(f = cal.CPO)

###############################################################################
###############################################################################

# Function to estimate the posterior comparison statistics

comp.stat <- function (chain, est, data, lc, lp) {
  
  D <- numeric(length = nrow(x = chain))
  
  for (i in 1:nrow(x = chain)) D[i] <- loglike(a = chain[i,], data = data)
  
  D.bar <- -2 * mean(x = D)
  D.hat <- -2 * loglike(a = colMeans(x = chain), data = data)
  
  DIC  <- 2 * D.bar - D.hat
  EAIC <- D.bar + 2 * ncol(x = chain)
  EBIC <- D.bar + ncol(x = chain) * log(x = data$n)
  LMPL <- sum(x = log(x = lc$CPO))
  pB   <- sim.pB(data = data, est = est, lp = lp)
  
  lapply(X = list(DIC = DIC, EAIC = EAIC, EBIC = EBIC, NLMPL = -LMPL, pB = pB), FUN = round, digits = 3)
  
}
comp.stat <- cmpfun(f = comp.stat)

###############################################################################
###############################################################################

# Function to estimate the posterior divergence measures

div.mea <- function (M, data, lc) {
  
  CPO <- lc$CPO
  aux <- lc$aux
  
  KL <- L1 <- CS <- H <- R.KL <- R.L1 <- R.CS <- R.H <- numeric(length = data$n)
  
  for (i in 1:data$n) {
    
    KL[i]   <- -log(x = CPO[i]) + sum(x = log(x = aux[i,]))/M
    R.KL[i] <- (1 + sqrt(x = 1 - exp(x = -2 * KL[i])))/2
    
    L1[i]   <- sum(x = abs(x = CPO[i] - aux[i,])/aux[i,])/(2 * M)
    R.L1[i] <- .5 + L1[i]
    
    CS[i]   <- sum(x = (CPO[i] - aux[i,])^2/aux[i,]^2)/M
    R.CS[i] <- (1 + sqrt(x = CS[i]))/2
    
    H[i]   <- sum(x = (sqrt(x = CPO[i]/aux[i,]) - 1)^2)/(2 * M)
    R.H[i] <- .5 + (sqrt(x = H[i]) - sqrt(x = H[i]^3)) * sqrt(x = 2 - H[i])
    
  }
  
  list(KL = KL, R.KL = R.KL, L1 = L1, R.L1 = R.L1, CS = CS, R.CS = R.CS, H = H, R.H = R.H)
  
}
div.mea <- cmpfun(f = div.mea)

###############################################################################
###############################################################################

# Sequential search algorithm to generate pseudo-random values from the ZMPS
# distribution

SeqSea.d <- function (a, b) {
  
  k <- 0
  u <- runif(n = 1)
  p <- dzmps(y = k, a = a, b = b)
  
  while (u > p) {
    
    k <- k + 1
    p <- p + dzmps(y = k, a = a, b = b)
    
  }
  
  k
  
}
SeqSea.d <- cmpfun(f = SeqSea.d)

###############################################################################
###############################################################################

# Monte Carlo estimator for the posterior predictive distribution (ppd)

dzmps.p <- function (w, a, b) {
  
  sum(x = sapply(X = w, FUN = dzmps, a = a, b = b))/length(x = a)
  
}
dzmps.p <- cmpfun(f = dzmps.p)

###############################################################################
###############################################################################

# Sequential Search algorithm to generate pseudo-random values from the ppd

SeqSea.p <- function (C1, C2) {
  
  k <- 0
  u <- runif(n = 1)
  p <- dzmps.p(w = k, a = C1, b = C2)
  
  while (u > p) {

    k <- k + 1
    p <- p + dzmps.p(w = k, a = C1, b = C2)
    
  }
  
  k
  
}
SeqSea.p <- cmpfun(f = SeqSea.p)

###############################################################################
###############################################################################

# Function to estimate Bayesian p-values (via simulation)

sim.pB <- function (data, est, lp, B = 500) {
  
  yp  <- numeric(length = data$n)
  aux <- numeric(length = B)
  
  eX <- lp$eX
  eV <- lp$eV

  for (i in 1:B) {
    
    for (j in 1:data$n) yp[j] <- SeqSea.p(C1 = exp(x = h(a = exp(x = eX[j,]), log = T)),
                                          C2 = make.link(link = data$link)$linkinv(eta = eV[j,]))
    
    aux[i] <- -2 * sum(x = dzmps(y = yp, a = exp(x = h(a = est[1], log = T)), b = est[2], log = T))

  }

  mean(x = aux >= (-2 * sum(x = dzmps(y = data$y, a = exp(x = h(a = est[1], log = T)), b = est[2], log = T))))
  
}
sim.pB <- cmpfun(f = sim.pB)

###############################################################################
###############################################################################

# Function to summarize the posterior distribution

post.sum <- function (chain, data, digits) {
  
  id.b1 <- 1:ncol(x = data$X1)
  id.b2 <- (max(x = id.b1) + 1):(ncol(x = data$X1) + ncol(x = data$X2))
  
  chain.mu <- chain[, id.b1]
  chain.om <- chain[, id.b2]
  
  mean.b1 <- rbind(a = mean(x = chain.mu[, 1]), b = mean(x = chain.mu[, 2]))
  medn.b1 <- rbind(a = median(x = chain.mu[, 1]), b = median(x = chain.mu[, 2]))
  std.b1  <- rbind(a = sd(x = chain.mu[, 1]), b = sd(x = chain.mu[, 2]))
  ess.b1  <- rbind(a = effectiveSize(x = chain.mu[, 1]), b = effectiveSize(x = chain.mu[, 2]))
  HPDI.b1 <- HPDI(x = chain.mu)
  
  mean.b2 <- rbind(a = mean(x = chain.om[, 1]), b = mean(x = chain.om[, 2]))
  medn.b2 <- rbind(a = median(x = chain.om[, 1]), b = median(x = chain.om[, 2]))
  std.b2  <- rbind(a = sd(x = chain.om[, 1]), b = sd(x = chain.om[, 2]))
  ess.b2  <- rbind(a = effectiveSize(x = chain.om[, 1]), b = effectiveSize(x = chain.om[, 2]))
  HPDI.b2 <- HPDI(x = chain.om)
  
  aux.tab <- cbind(a = rbind(a = mean.b1, b = mean.b2),
                   b = rbind(a = medn.b1, b = medn.b2),
                   c = rbind(a = std.b1, b = std.b2),
                   d = rbind(a = ess.b1, b = ess.b2),
                   e = rbind(a = HPDI.b1, b = HPDI.b2))
  
  rownames(x = aux.tab) <- c('beta10', 'beta11', 'beta20', 'beta21')
  colnames(x = aux.tab) <- c('Mean', 'Median', 'Std. Dev.', 'ESS', 'Lower', 'Upper')
  
  apply(X = aux.tab, MARGIN = 2, FUN = round, digits = digits)
  
}
post.sum <- cmpfun(f = post.sum)

###############################################################################
###############################################################################

# Function to estimate the Highest Posterior Density Intervals

HPDI <- function (x, cred.int = .95) {
  
  x <- as.matrix(x = x)
  v <- apply(X = x, MARGIN = 2, FUN = sort)
  M <- nrow(x = v)
  k <- ncol(x = v)
  
  gap <- max(x = 1, y = min(x = M - 1, y = round(x = M * cred.int)))
  ini <- 1:(M - gap)
  id  <- apply(X = v[ini + gap, , drop = F] - v[ini, , drop = F], MARGIN = 2, FUN = which.min)
  
  aux <- cbind(a = v[cbind(a = id, b = 1:k)], b = v[cbind(a = id + gap, b = 1:k)])
  
  colnames(x = aux) <- c('Lower', 'Upper')
  
  aux
  
}
HPDI <- cmpfun(f = HPDI)

###############################################################################
###############################################################################