###############################################################################
###############################################################################
# PROGRAM:
# fit-zmps-model.R
# 
# USE:
# Fit the ZMPS model
#     
# AUTHOR:
# Wesley Bertoli (wbsilva@utfpr.edu.br)
#
# DATE:
# January 22, 2021
###############################################################################
###############################################################################

# Clear workspace

rm(list = ls())

###############################################################################
###############################################################################

# Loading required packages

invisible(x = lapply(X = c('coda', 'compiler', 'MASS', 'mvtnorm', 'numDeriv'), FUN = require, character.only = T))

###############################################################################
###############################################################################

# Enabling just-in-time compilation

enableJIT(level = 3)

###############################################################################
###############################################################################

# Define the required functions in the workspace

source(file = 'Codes/auxiliar-functions.R')

###############################################################################
###############################################################################

# Inputting data

db <- read.table(file = 'Data/data_chr_aberration.txt', header = T, sep = ';')

# Regression matrix

X <- cbind(a = 1, b = db$x)

# Creating a list of working objects

aux.data <- list(n = nrow(x = db), y = db$y, X1 = X, X2 = X, link = 'probit')

###############################################################################
###############################################################################

# Model fit and inference

set.seed(seed = 666)

## Generating chains

fit <- RwM(data = aux.data, print.it = T)

## Obtaining the final chains

Q <- fit$chain[seq(from = 1, to = nrow(x = fit$chain), by = 10),]

# Posterior parameter estimates and 95% HPDIs from ZMPS fitted model

table3 <- post.sum(chain = Q, data = aux.data, digits = 3)

## Obtaining overall estimates under the classical formulation of ZM models

est <- or.pars(chain = Q, data = aux.data)

## Estimating the linear predictors

lp <- lin.pred(chain = Q, data = aux.data)

## Estimating the Conditional Predictive Ordinate

lc <- cal.CPO(M = nrow(x = Q), data = aux.data, lp = lp)

## Estimating the posterior comparison statistics

cmp <- comp.stat(chain = Q, est = est, data = aux.data, lc = lc, lp = lp)

## Estimating the posterior divergence measures

div <- div.mea(M = nrow(x = Q), data = aux.data, lc = lc)

###############################################################################
###############################################################################

# Saving the results

save.image(file = 'RData/zmps_fitted_model.RData')

###############################################################################
###############################################################################