###############################################################################
###############################################################################
# PROGRAM:
# rqr-specific-functions.R
# 
# USE:
# Define specific functions to be used when estimating the Randomized Quantile
# Residuals (RQRs)
#     
# AUTHOR:
# Wesley Bertoli (wbsilva@utfpr.edu.br)
#
# DATE:
# January 22, 2021
###############################################################################
###############################################################################

# Cumulative mass function of the ZMPS distribution
  
pzmps <- function (q, a, b) {
  
  1 - b * (a^4 + a^3 * q + a^2 * q^2 + 4 * a^3 + 6 * a^2 * q + 10 * a^2 + 2 * a * q + 7 * a + 2)/
          ((a + 1)^(q + 3) * (a^2 + a + 2) * (1 - (a^5 + 3 * a^4 + 4 * a^3)/((a^2 + a + 2) * (a + 1)^3)))
  
}
pzmps <- cmpfun(f = pzmps)

###############################################################################
###############################################################################

# Function to compute the RQRs

rqr.fun <- function (y, a, b) {
  
  qnorm(p = runif(n   = length(x = y),
                  min = ifelse(test = y > 0, yes = pzmps(q = y - 1, a = exp(x = h(a = a, log = T)), b = b), no = 0),
                  max = pzmps(q = y, a = exp(x = h(a = a, log = T)), b = b)))
  
}
rqr.fun <- cmpfun(f = rqr.fun)

###############################################################################

# Main function

to.run <- function (i, data, lp) {
  
  err <- T
  
  while (err) {
    
    theta <- exp(x = h(a = exp(x = rowMeans(x = lp$eX)), log = T))
    omega <- make.link(link = data$link)$linkinv(eta = rowMeans(x = lp$eV))
      
    for (j in 1:data$n) data$y[j] <- SeqSea.d(a = theta[j], b = omega[j])

    fit <- try(expr = RwM(data = data), silent = T)
    err <- ifelse(test = !inherits(x = fit, what = 'try-error'), yes = F, no = T)

  }
  
  lp.new <- lin.pred(chain = fit$chain[seq(from = 1, to = nrow(x = fit$chain), by = 10),], data = data)
  est.mu <- exp(x = rowMeans(x = lp.new$eX))
  est.om <- make.link(link = data$link)$linkinv(eta = rowMeans(x =  lp.new$eV))
  
  list(rqr = rqr.fun(y = data$y, a = est.mu, b = est.om))
  
}
to.run <- cmpfun(f = to.run)

###############################################################################
###############################################################################