###############################################################################
###############################################################################
# PROGRAM:
# simulation-last-code.R
# 
# USE:
# Provide the final posterior summary for the simulation study based on the
# ZMPS model
#     
# AUTHOR:
# Wesley Bertoli (wbsilva@utfpr.edu.br)
#
# DATE:
# January 22, 2021
###############################################################################
###############################################################################

# Function to return the final posterior summary

final.sum <- function (M, chain, parms) {
  
  b0  <- parms[1]
  b1  <- parms[2]
  g0  <- parms[3]
  g1  <- parms[4]
  
  aux1.b0 <- aux1.b1 <- aux1.g0 <- aux1.g1 <-
  aux2.b0 <- aux2.b1 <- aux2.g0 <- aux2.g1 <-
  bs1.b0  <- bs1.b1  <- bs1.g0  <- bs1.g1  <-
  qe1.b0  <- qe1.b1  <- qe1.g0  <- qe1.g1  <-
  mp1.b0  <- mp1.b1  <- mp1.g0  <- mp1.g1  <-
  id.inf.HPD.b0 <- id.inf.HPD.b1 <- id.inf.HPD.g0 <- id.inf.HPD.g1 <-
  id.sup.HPD.b0 <- id.sup.HPD.b1 <- id.sup.HPD.g0 <- id.sup.HPD.g1 <- numeric(length = M)
  
  ci.HPD.b0 <- ci.HPD.b1 <- ci.HPD.g0 <- ci.HPD.g1 <- array(dim = c(M, 2))
  
  for (i in 1:M) {
      
    chain.b0 <- chain[[i]][[1]][, 1]
    chain.b1 <- chain[[i]][[1]][, 2]
    chain.g0 <- chain[[i]][[1]][, 3]
    chain.g1 <- chain[[i]][[1]][, 4]
    
    aux1.b0[i] <- mean(x = chain.b0)
    aux1.b1[i] <- mean(x = chain.b1)
    aux1.g0[i] <- mean(x = chain.g0)
    aux1.g1[i] <- mean(x = chain.g1)
    
    aux2.b0[i] <- median(x = chain.b0)
    aux2.b1[i] <- median(x = chain.b1)
    aux2.g0[i] <- median(x = chain.g0)
    aux2.g1[i] <- median(x = chain.g1)
    
    bs1.b0[i] <- aux1.b0[i] - b0
    bs1.b1[i] <- aux1.b1[i] - b1
    bs1.g0[i] <- aux1.g0[i] - g0
    bs1.g1[i] <- aux1.g1[i] - g1
    
    qe1.b0[i] <- bs1.b0[i]^2
    qe1.b1[i] <- bs1.b1[i]^2
    qe1.g0[i] <- bs1.g0[i]^2
    qe1.g1[i] <- bs1.g1[i]^2
    
    mp1.b0[i] <- abs(x = (aux1.b0[i] - b0)/b0)
    mp1.b1[i] <- abs(x = (aux1.b1[i] - b1)/b1)
    mp1.g0[i] <- abs(x = (aux1.g0[i] - g0)/g0)
    mp1.g1[i] <- abs(x = (aux1.g1[i] - g1)/g1)
    
    HPD.b1 <- HPDI(x = cbind(a = chain.b0, b = chain.b1))
    HPD.b2 <- HPDI(x = cbind(a = chain.g0, b = chain.g1))
    
    ci.HPD.b0[i,] <- HPD.b1[1,]
    ci.HPD.b1[i,] <- HPD.b1[2,]
    ci.HPD.g0[i,] <- HPD.b2[1,]
    ci.HPD.g1[i,] <- HPD.b2[2,]
    
    id.inf.HPD.b0[i] <- ifelse(test = b0 < min(x = ci.HPD.b0[i,]), yes = 1, no = 0)
    id.inf.HPD.b1[i] <- ifelse(test = b1 < min(x = ci.HPD.b1[i,]), yes = 1, no = 0)
    id.inf.HPD.g0[i] <- ifelse(test = g0 < min(x = ci.HPD.g0[i,]), yes = 1, no = 0)
    id.inf.HPD.g1[i] <- ifelse(test = g1 < min(x = ci.HPD.g1[i,]), yes = 1, no = 0)
    
    id.sup.HPD.b0[i] <- ifelse(test = b0 > max(x = ci.HPD.b0[i,]), yes = 1, no = 0)
    id.sup.HPD.b1[i] <- ifelse(test = b1 > max(x = ci.HPD.b1[i,]), yes = 1, no = 0)
    id.sup.HPD.g0[i] <- ifelse(test = g0 > max(x = ci.HPD.g0[i,]), yes = 1, no = 0)
    id.sup.HPD.g1[i] <- ifelse(test = g1 > max(x = ci.HPD.g1[i,]), yes = 1, no = 0)
    
  }
  
  mean.b0 <- mean(x = aux1.b0)
  mean.b1 <- mean(x = aux1.b1)
  mean.g0 <- mean(x = aux1.g0)
  mean.g1 <- mean(x = aux1.g1)
  
  medn.b0 <- mean(x = aux2.b0)
  medn.b1 <- mean(x = aux2.b1)
  medn.g0 <- mean(x = aux2.g0)
  medn.g1 <- mean(x = aux2.g1)
  
  bias1.b0 <- mean(x = bs1.b0)
  bias1.b1 <- mean(x = bs1.b1)
  bias1.g0 <- mean(x = bs1.g0)
  bias1.g1 <- mean(x = bs1.g1)
  
  mse1.b0 <- mean(x = qe1.b0)
  mse1.b1 <- mean(x = qe1.b1)
  mse1.g0 <- mean(x = qe1.g0)
  mse1.g1 <- mean(x = qe1.g1)
  
  sd1.b0 <- sqrt(x = sum(x = aux1.b0^2)/M - mean.b0^2)
  sd1.b1 <- sqrt(x = sum(x = aux1.b1^2)/M - mean.b1^2)
  sd1.g0 <- sqrt(x = sum(x = aux1.g0^2)/M - mean.g0^2)
  sd1.g1 <- sqrt(x = sum(x = aux1.g1^2)/M - mean.g1^2)
  
  smse1.var.b0 <- sqrt(x = mse1.b0)/sd1.b0
  smse1.var.b1 <- sqrt(x = mse1.b1)/sd1.b1
  smse1.var.g0 <- sqrt(x = mse1.g0)/sd1.g0
  smse1.var.g1 <- sqrt(x = mse1.g1)/sd1.g1
  
  mape1.b0 <- mean(x = mp1.b0) * 100
  mape1.b1 <- mean(x = mp1.b1) * 100
  mape1.g0 <- mean(x = mp1.g0) * 100
  mape1.g1 <- mean(x = mp1.g1) * 100
  
  cp.inf.HPD.b0 <- sum(x = id.inf.HPD.b0)/M * 100
  cp.inf.HPD.b1 <- sum(x = id.inf.HPD.b1)/M * 100
  cp.inf.HPD.g0 <- sum(x = id.inf.HPD.g0)/M * 100
  cp.inf.HPD.g1 <- sum(x = id.inf.HPD.g1)/M * 100
  
  cp.sup.HPD.b0 <- sum(x = id.sup.HPD.b0)/M * 100
  cp.sup.HPD.b1 <- sum(x = id.sup.HPD.b1)/M * 100
  cp.sup.HPD.g0 <- sum(x = id.sup.HPD.g0)/M * 100
  cp.sup.HPD.g1 <- sum(x = id.sup.HPD.g1)/M * 100
  
  cp.in.HPD.b0 <- 100 - cp.inf.HPD.b0 - cp.sup.HPD.b0
  cp.in.HPD.b1 <- 100 - cp.inf.HPD.b1 - cp.sup.HPD.b1
  cp.in.HPD.g0 <- 100 - cp.inf.HPD.g0 - cp.sup.HPD.g0
  cp.in.HPD.g1 <- 100 - cp.inf.HPD.g1 - cp.sup.HPD.g1
  
  cif.HPD.b0 <- colMeans(x = ci.HPD.b0)
  cif.HPD.b1 <- colMeans(x = ci.HPD.b1)
  cif.HPD.g0 <- colMeans(x = ci.HPD.g0)
  cif.HPD.g1 <- colMeans(x = ci.HPD.g1)
  
  round(x = c(mean.b0, mean.b1, mean.g0, mean.g1,
              medn.b0, medn.b1, medn.g0, medn.g1,
              sd1.b0, sd1.b1, sd1.g0, sd1.g1,
              cif.HPD.b0[1], cif.HPD.b1[1], cif.HPD.g0[1], cif.HPD.g1[1],
              cif.HPD.b0[2], cif.HPD.b1[2], cif.HPD.g0[2], cif.HPD.g1[2],
              bias1.b0, bias1.b1, bias1.g0, bias1.g1,
              mse1.b0, mse1.b1, mse1.g0, mse1.g1,
              smse1.var.b0, smse1.var.b1, smse1.var.g0, smse1.var.g1,
              mape1.b0, mape1.b1, mape1.g0, mape1.g1,
              cp.inf.HPD.b0, cp.inf.HPD.b1, cp.inf.HPD.g0, cp.inf.HPD.g1,
              cp.in.HPD.b0, cp.in.HPD.b1, cp.in.HPD.g0, cp.in.HPD.g1,
              cp.sup.HPD.b0, cp.sup.HPD.b1, cp.sup.HPD.g0, cp.sup.HPD.g1), digits = 4)
  
}

summary <- array(dim = c(48, 1))
colnames(x = summary) <- paste('n', '=', n)

rownames(x = summary) <-
c('mean.b0', 'mean.b1', 'mean.g0', 'mean.g1',
  'medn.b0', 'medn.b1', 'medn.g0', 'medn.g1',
  'sd.b0', 'sd.b1', 'sd.g0', 'sd.g1',
  'HPD.b0.low', 'HPD.b1.low', 'HPD.g0.low', 'HPD.g1.low',
  'HPD.b0.upp', 'HPD.b1.upp', 'HPD.g0.upp', 'HPD.g1.upp',
  'bias1.b0', 'bias1.b1', 'bias1.g0', 'bias1.g1',
  'mse1.b0', 'mse1.b1', 'mse1.g0', 'mse1.g1',
  'smse1.var.b0', 'smse1.var.b1', 'smse1.var.g0', 'smse1.var.g1',
  'mape1.b0', 'mape1.b1', 'mape1.g0', 'mape1.g1',
  'cp.inf.HPD.b0', 'cp.inf.HPD.b1', 'cp.inf.HPD.g0', 'cp.inf.HPD.g1',
  'cp.in.HPD.b0', 'cp.in.HPD.b1', 'cp.in.HPD.g0', 'cp.in.HPD.g1',
  'cp.sup.HPD.b0', 'cp.sup.HPD.b1', 'cp.sup.HPD.g0', 'cp.sup.HPD.g1')

final.sum <- cmpfun(f = final.sum)

summary[, 1] <- final.sum(M = M, chain = result, parms = parms)

###############################################################################
###############################################################################