###############################################################################
###############################################################################
# PROGRAM:
# simulation-zmps-model.R
# 
# USE:
# Simulation study for the ZMPS model
#     
# AUTHOR:
# Wesley Bertoli (wbsilva@utfpr.edu.br)
#
# DATE:
# January 22, 2021
###############################################################################
###############################################################################

# Clear workspace

rm(list = ls())

###############################################################################
###############################################################################

# Loading required packages

invisible(x = lapply(X = c('compiler', 'parallel'), FUN = require, character.only = T))

###############################################################################
###############################################################################

# Enabling just-in-time compilation

enableJIT(level = 3)

###############################################################################
###############################################################################

# Starting a cluster with specific number of cores

n.cores <- detectCores() - 1

cl <- makeCluster(spec = n.cores)

###############################################################################
###############################################################################

# Export required packages to cluster

invisible(x = clusterEvalQ(cl   = cl,
                           expr = { lapply(X   = c('MASS', 'mvtnorm', 'numDeriv'),
                                           FUN = require, character.only = T) }))

###############################################################################
###############################################################################

# Define the required functions in the workspace

source(file = 'Codes/auxiliar-functions.R')
source(file = 'Codes/auxiliar-functions-simulation.R')

###############################################################################
###############################################################################

# Exporting functions to cluster

fun.list <- c('dps', 'dztps', 'dzmps', 'h', 'loglike.b1', 'loglike.b2', 'logpost.b1',
              'logpost.b2', 'makePD', 'RwM', 'HPDI', 'SeqSea.d', 'gen.parms', 'to.run')

clusterExport(cl = cl, varlist = fun.list)

###############################################################################
###############################################################################

# Simulation setup

case   <- 1 # 1 or 2 (1 = Inflation, 2 = Deflation)
sc     <- 1 # 1-4

link  <- 'logit' # logit, probit, or cloglog
path1 <- paste0(path, '/RData/simul_n')

obj.list <- c('cl', 'case', 'sc', 'link', 'path1')

###############################################################################
###############################################################################

# Simulation and estimation

## n = 50

n <- 50

source(file = 'Codes/simulation-first-code.R')
source(file = 'Codes/simulation-last-code.R')
rm(list = setdiff(x = ls(), y = c(fun.list, obj.list, 'fun.list', 'obj.list', 'n', 'summary')))
save(list = c('case', 'sc', 'n', 'link', 'summary'),
     file = paste0(path1, n, '_case', case, '_sc', sc, '_', link, '.RData'))

## n = 100

n <- 100

source(file = 'Codes/simulation-first-code.R')
source(file = 'Codes/simulation-last-code.R')
rm(list = setdiff(x = ls(), y = c(fun.list, obj.list, 'fun.list', 'obj.list', 'n', 'summary')))
save(list = c('case', 'sc', 'n', 'link', 'summary'),
     file = paste0(path1, n, '_case', case, '_sc', sc, '_', link, '.RData'))

## n = 200

n <- 200

source(file = 'Codes/simulation-first-code.R')
source(file = 'Codes/simulation-last-code.R')
rm(list = setdiff(x = ls(), y = c(fun.list, obj.list, 'fun.list', 'obj.list', 'n', 'summary')))
save(list = c('case', 'sc', 'n', 'link', 'summary'),
     file = paste0(path1, n, '_case', case, '_sc', sc, '_', link, '.RData'))

## n = 500

n <- 500

source(file = 'Codes/simulation-first-code.R')
source(file = 'Codes/simulation-last-code.R')
rm(list = setdiff(x = ls(), y = c(fun.list, obj.list, 'fun.list', 'obj.list', 'n', 'summary')))
save(list = c('case', 'sc', 'n', 'link', 'summary'),
     file = paste0(path1, n, '_case', case, '_sc', sc, '_', link, '.RData'))

stopCluster(cl = cl)

###############################################################################
###############################################################################